//
//		Plugin DLL for Force-Sender program out of the X-Sim packet
//		Game: Richard Burns Rally
//
//		Thanks go to: 
//		- Vladimir Kadlec (base addresses)
//		- Nye (many offsets)
//		- oleroc
//		- jamesdio (for speed offsets in the injector)
//		- Dave Skotnicki (most active RBR developer and the source of jamesdio)
//
//		And for continous support of the RBR thread:
//		- tronicgr
//		- egoexpress
//

#include "stdafx.h"
#include <math.h>
#define M_PI       3.14159265358979323846

#pragma data_seg ("rbrplugin")
float oldracetime=0;
int outcount=0;
BOOL rbr_detected=FALSE;
int scandelay=0;
DWORD procid=0;
double oldspeedx=0;
double oldspeedy=0;
double oldspeedz=0;
#pragma data_seg()
#pragma comment(linker,"/SECTION:rbrplugin,RWS")

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#ifdef _MSC_VER
  #pragma pack(push, 1)
#else
  #pragma pack(1)
#endif

#ifdef _MSC_VER
  #pragma pack(pop)
#else
  #pragma pack()
#endif

#define DLLEXPORT extern "C" __declspec (dllexport)

//This structure will begin with the ingame memory address stored at 23460968
struct RBRGaugeInfo	//Offset collection of Vladimir Kadlec and NYE (known from GTR PDMonitor series)
{
	char pad0[0xC];			//													First offset of base pointer with 12 bytes
	float speed;			//0x0C  - 0x0F		gauge speed in m/s
	float rpm;				//0x10  - 0x13		engine rounds per minute
	float temp;				//0x14  - 0x17		engine temperature
	float turbo;			//0x18  - 0x1B		in Pascals
	char pad1[0x4];			//0x1C  - 0x1F										Offset with   4 bytes
	float way_after;		//0x20  - 0x23		Distance from start
	char pad2[0x4];			//0x24  - 0x27										Offset with   4 bytes
	float way_to_do;		//0x28  - 0x2B		Distance to finish
	char pad3[0x114];		//0x2C  - 0x13F										Offset with 276 bytes
    float racetime;			//0x140 - 0x143		racetime in ms
	char pad4[0x18];		//0x144 - 0x15B										Offset with  24 bytes
	float sessiontime;		//0x15C - 0x15F		SessionTime 
	char pad5[0x10];		//0x160 - 0x16F										Offset with  16 bytes
	char gear_number;		//0x170 - 0x171		current gear number (2= first gear)
};

//This structure will begin with the ingame memory address stored at 9369184
struct RBRCarInfo  //Checked in to x-sim force injector by Jason Jamesdio and from the source code of RBRNet of Dave Skotnicki
{
	char pad0[256];
	float car_look_at[4];
	char pad1[128];
	float map_rotation_speed_x;
	float map_rotation_speed_y;
	float map_rotation_speed_z;
	char pad2[36];
	float map_speed_x;
	float map_speed_y;
	float map_speed_z;
};

struct GetPluginData
{
	int			data[22];
};

struct GetPluginDataEx
{
	int			data[22];				//Simulator plugin values
	int			speed;					//current speed in kmh, mph or m/s
	int			rpm;					//rpm
	int			fuel;					//fuel of tank
	int			tempfuel;				//tank temperature
	int			currentgear;			//number of gear
	int			currentlap;				//number of current lab
	int			bestlaptime;			//best lap time
	int			lastlaptime;			//last lap time
	int			position;				//current position
	int			tempwater;				//water temperature
	int			flags;					//flags
	int			tiretemps[4][3];		//temp of each tire (inner middle outer)
	int			altimeter;				//Flight gauge altimeter
	int			airspeed;				//Flight gauge airspeed over ground
	int			verticalvelocity;		//Flight Vertical speed meter
	int			attidude;				//Flight attitude Indicator
	int			heading;				//Flight heading indicator
	int			fuelflow;				//Flight fuelflow
	int			fuelleft;				//Flight fuelleft
	int			fuelright;				//Flight fuelright
	int			tempexhaustgas;			//Flight temperature of exhaust gas
	int			oiltemperature;			//Flight temperature of oil
	int			oilpressure;			//Flight pressure of oil
	int			turncoordinator;		//Flight turncoordinator
	int			outsidetemperture;		//Flight outside air temperature
	int			ampmeter;				//Flight ampmeter
	int			buffer[6];				//not defined values left
};

struct GetInfo
{
	char		pluginname[51];
	char		plugininfo[20][31];
};

struct DllInfo
{
	char		type[11];
	char		name[11];
};

struct AutoScanInfo
{
	int			numberofentries;
	char		findwindowsname[20][260];
	char		executablename[20][260];
};


DLLEXPORT DllInfo GetType(int number);
DLLEXPORT GetInfo GetPluginName(int number);
DLLEXPORT GetPluginData GetDataArray(int number);
DLLEXPORT GetPluginDataEx GetDataArrayEx(int number);
DLLEXPORT AutoScanInfo GetAutoScanInfo(int number);

BOOL IsPIDActive(DWORD fpid)
{
	HWND h = ::GetTopWindow(0 );
	DWORD pid;
	while ( h )
	{
		pid=0;
		DWORD dwTheardId = ::GetWindowThreadProcessId( h,&pid);
		if ( pid == fpid )
		{
			return TRUE;
		}
		h = ::GetNextWindow( h , GW_HWNDNEXT);
	}
	return FALSE;
}

BOOL ScanForGame(void)
{
	if(rbr_detected==TRUE)
	{
		rbr_detected=IsPIDActive(procid);
		if(rbr_detected==FALSE)
		{
			oldracetime=0;
			outcount=0;
			oldspeedx=0;
			oldspeedy=0;
			oldspeedz=0;
		}
		return rbr_detected;
	}
	scandelay++;
	if(scandelay > 10)
	{
		scandelay =0;
		HWND FindWnd=NULL;
		HANDLE game_process_handle=NULL;
		FindWnd = ::FindWindowA(0,"Richard Burns Rally - DirectX9");
		if(FindWnd)
		{ 
			GetWindowThreadProcessId(FindWnd, &procid);
			game_process_handle = OpenProcess(PROCESS_ALL_ACCESS, FALSE, procid);
			if(game_process_handle)
			{
				CloseHandle(game_process_handle);
				rbr_detected=TRUE;
				oldracetime=0;
				outcount=0;
				oldspeedx=0;
				oldspeedy=0;
				oldspeedz=0;
				return TRUE;
			}
			return FALSE;
		}
		return FALSE;
	}
	return FALSE;
}

AutoScanInfo GetAutoScanInfo(int number)
{
	//This are the different auto detection names for the sender (window name and exe name)
	AutoScanInfo newinfo={0};
	CString windowname="Richard Burns Rally - DirectX9";
	for(int z=0; z < windowname.GetLength(); z++){newinfo.findwindowsname[0][z]=windowname[z];}
	CString programname="RichardBurnsRally.exe";
	for(int z=0; z < programname.GetLength(); z++){newinfo.executablename[0][z]=programname[z];}
	windowname="Richard Burns Rally - DirectX9";
	for(int z=0; z < windowname.GetLength(); z++){newinfo.findwindowsname[1][z]=windowname[z];}
	programname="RichardBurnsRally_NoSSE.exe";
	for(int z=0; z < programname.GetLength(); z++){newinfo.executablename[1][z]=programname[z];}
	windowname="Richard Burns Rally - DirectX9";
	for(int z=0; z < windowname.GetLength(); z++){newinfo.findwindowsname[2][z]=windowname[z];}
	programname="RichardBurnsRally_SSE.exe";
	for(int z=0; z < programname.GetLength(); z++){newinfo.executablename[2][z]=programname[z];}
	newinfo.numberofentries=3;
	return newinfo;
}

DllInfo GetType(int number)
{
	DllInfo dllinfo;
	char* type;
	type="SenderPlug";
	for(int z=0; z < 10; z++){dllinfo.type[z]=type[z];}
	dllinfo.type[10]= 0;
	char* name;
	name="RBR Plugin";
	for(int z=0; z < 10; z++){dllinfo.name[z]=name[z];}
	dllinfo.name[10]= 0;
	return dllinfo;
}

GetInfo GetPluginName(int number)
{
	GetInfo getinfo;
	char* name;
	// name of plugin, must be 50 characters !!	(0 terminated)
	name="Richard Burns Rally plugin                          ";
	for(int z=0; z < 50; z++){getinfo.pluginname[z]=name[z];}
	// name of each effect
	// descriptions must be 30 character long (0 terminated)
	char* description1 ="Effect 21: Current gear       ";
	char* description2 ="Effect 22: Current speed      ";
	char* description3 ="Effect 23: nothing            ";
	char* description4 ="Effect 24: nothing            ";
	char* description5 ="Effect 25: Lateral force      ";
	char* description6 ="Effect 26: Vertical force     ";
	char* description7 ="Effect 27: Longitudinal force ";
	char* description8 ="Effect 28: Pitch              ";
	char* description9 ="Effect 29: Yaw                ";
	char* description10="Effect 30: Roll               ";
	char* description11="Effect 31: car speed x        ";
	char* description12="Effect 32: car speed y        ";
	char* description13="Effect 33: car speed z        ";
	char* description14="Effect 34: nothing            ";
	char* description15="Effect 35: nothing            ";
	char* description16="Effect 36: nothing            ";
	char* description17="Effect 37: nothing            ";
	char* description18="Effect 38: nothing            ";
	char* description19="Effect 39: race time          ";
	char* description20="Effect 40: session time       ";
	// copy in array
	for(int z=0; z < 30; z++){getinfo.plugininfo[0][z]=description1[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[1][z]=description2[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[2][z]=description3[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[3][z]=description4[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[4][z]=description5[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[5][z]=description6[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[6][z]=description7[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[7][z]=description8[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[8][z]=description9[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[9][z]=description10[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[10][z]=description11[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[11][z]=description12[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[12][z]=description13[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[13][z]=description14[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[14][z]=description15[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[15][z]=description16[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[16][z]=description17[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[17][z]=description18[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[18][z]=description19[z];}
	for(int z=0; z < 30; z++){getinfo.plugininfo[19][z]=description20[z];}
	// zero termination of strings
	getinfo.pluginname[50]=0;
	getinfo.plugininfo[0][30]=0;
	getinfo.plugininfo[1][30]=0;
	getinfo.plugininfo[2][30]=0;
	getinfo.plugininfo[3][30]=0;
	getinfo.plugininfo[4][30]=0;
	getinfo.plugininfo[5][30]=0;
	getinfo.plugininfo[6][30]=0;
	getinfo.plugininfo[7][30]=0;
	getinfo.plugininfo[8][30]=0;
	getinfo.plugininfo[9][30]=0;
	getinfo.plugininfo[10][30]=0;
	getinfo.plugininfo[11][30]=0;
	getinfo.plugininfo[12][30]=0;
	getinfo.plugininfo[13][30]=0;
	getinfo.plugininfo[14][30]=0;
	getinfo.plugininfo[15][30]=0;
	getinfo.plugininfo[16][30]=0;
	getinfo.plugininfo[17][30]=0;
	getinfo.plugininfo[18][30]=0;
	getinfo.plugininfo[19][30]=0;
	//overgive info array
	return getinfo;
}

GetPluginData GetDataArray(int number) //Will be called by the old sender of X-Sim 1.6.x with the polling speed of the cluster dialog
{
	GetPluginData outarray={0};
	BOOL active=ScanForGame();   //Poll for the game and get the process ID
	if(active==TRUE)
	{
		if(procid != 0)
		{
			HANDLE game_process_handle=NULL;
			game_process_handle = OpenProcess(PROCESS_ALL_ACCESS, FALSE, procid);
			if(game_process_handle)
			{
				DWORD rw=0;
				unsigned long baseaddress=0;
				if(ReadProcessMemory(game_process_handle, (LPCVOID)23460968, &baseaddress, sizeof(unsigned long), &rw))	//Base memory location where the next memory address is stored
				{
					if(rw==sizeof(unsigned long) && baseaddress != 0 && baseaddress < 0xFFFFFF00)
					{
						RBRGaugeInfo gaugeinfo={0};
						RBRCarInfo carinfo={0};
						//Read out the gauge data
						DWORD sizeofstruct=sizeof(RBRGaugeInfo);
						if(ReadProcessMemory(game_process_handle, (LPCVOID)baseaddress, &gaugeinfo, sizeofstruct, &rw))
						{
							if(rw == sizeofstruct)
							{
								//Read out the car information
								if(ReadProcessMemory(game_process_handle, (LPCVOID)9369184, &baseaddress, sizeof(unsigned long), &rw)) //Base memory location where the next memory address is stored
								{
									if(rw==sizeof(unsigned long) && baseaddress != 0 && baseaddress < 0xFFFFFF00)
									{
										sizeofstruct=sizeof(RBRCarInfo);
										if(ReadProcessMemory(game_process_handle, (LPCVOID)baseaddress, &carinfo, sizeofstruct, &rw))
										{
											if(rw == sizeofstruct)
											{
												//Check if the game is in pause mode or timed out for some reason
												if(oldracetime == gaugeinfo.racetime)
												{
													if(outcount > 4)
													{
														outarray.data[21]=1;
													}
													else
													{
														outcount++;
													}
												}
												else
												{
													outcount=0;
												}
												//calculate time from previous readout to now in s
												double timeleft=0;
												if(oldracetime != 0 && gaugeinfo.racetime != 0)
												{
													timeleft=gaugeinfo.racetime-oldracetime;
													if(timeleft < 0){timeleft=0;}
												}

												//Convert to Euler degrees (wikipedia: http://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles)
												float rad_pitch	= atan2(2*((carinfo.car_look_at[0]*carinfo.car_look_at[3])+(carinfo.car_look_at[1]*carinfo.car_look_at[2])),1-(2*((carinfo.car_look_at[0]*carinfo.car_look_at[0])+(carinfo.car_look_at[1]*carinfo.car_look_at[1]))));
												float rad_roll  =  asin(2*((carinfo.car_look_at[0]*carinfo.car_look_at[2])-(carinfo.car_look_at[3]*carinfo.car_look_at[1])));
												float rad_yaw	= atan2(2*((carinfo.car_look_at[0]*carinfo.car_look_at[1])+(carinfo.car_look_at[2]*carinfo.car_look_at[3])),1-(2*((carinfo.car_look_at[1]*carinfo.car_look_at[1])+(carinfo.car_look_at[2]*carinfo.car_look_at[2]))));

												//2D transformation of speeds (FAQ of www.x-simulator.de)
												float speedz = -sin(rad_yaw) * carinfo.map_speed_x + cos(rad_yaw) * carinfo.map_speed_y;
												float speedy = carinfo.map_speed_z;
												float speedx =  cos(rad_yaw) * carinfo.map_speed_x + sin(rad_yaw) * carinfo.map_speed_y;

												//Calculate accelerations out of change of speed
												float accelx = ((double(speedx)*1000000.000) - oldspeedx)*timeleft;
												float accely = ((double(speedy)*1000000.000) - oldspeedy)*timeleft;
												float accelz = ((double(speedz)*1000000.000) - oldspeedz)*timeleft;
												oldspeedx=double(speedx)*1000000.000;
												oldspeedy=double(speedy)*1000000.000;
												oldspeedz=double(speedz)*1000000.000;

												//Store current time
												oldracetime=gaugeinfo.racetime;
												//collect the outarray package
												if(outarray.data[21]==0) //game is active
												{
													//Transfer the simulator data
													outarray.data[1 ]=int(gaugeinfo.gear_number);
													outarray.data[2 ]=int(gaugeinfo.speed);

													//Car oriented accelerations
													outarray.data[5]=int(accelx*-1000.000);
													outarray.data[6]=int(accely*-1000.000);
													outarray.data[7]=int(accelz*-1000.000);

													//Output the car orientation in degree
													outarray.data[8 ]=int(rad_pitch*-(180.000/M_PI));
													float yawdegree=rad_yaw*(180.000/M_PI);
													if(yawdegree < 0){yawdegree=360.00+yawdegree;}
													outarray.data[9 ]=int(yawdegree);
													outarray.data[10]=int(rad_roll*-(180.000/M_PI));

													//Car oriented speeds
													outarray.data[11]=int(speedx*-1000.000);
													outarray.data[12]=int(speedy*-1000.000);
													outarray.data[13]=int(speedz*-1000.000);

													//Ingame times
													outarray.data[19]=int(gaugeinfo.racetime*1000.000);
													outarray.data[20]=int(gaugeinfo.sessiontime*1000.000);
													return outarray;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	//Failed
	outarray.data[21]=1;
	return outarray;
}

GetPluginDataEx GetDataArrayEx(int number) //Will be called by the sender 2.0 with the polling speed of the cluster dialog
{
	GetPluginDataEx outarray={0};
	BOOL active=ScanForGame();   //Poll for the game and get the process ID
	if(active==TRUE)
	{
		if(procid != 0)
		{
			HANDLE game_process_handle=NULL;
			game_process_handle = OpenProcess(PROCESS_ALL_ACCESS, FALSE, procid);
			if(game_process_handle)
			{
				DWORD rw=0;
				unsigned long baseaddress=0;
				if(ReadProcessMemory(game_process_handle, (LPCVOID)23460968, &baseaddress, sizeof(unsigned long), &rw))	//Base memory location where the next memory address is stored
				{
					if(rw==sizeof(unsigned long) && baseaddress != 0 && baseaddress < 0xFFFFFF00)
					{
						RBRGaugeInfo gaugeinfo={0};
						RBRCarInfo carinfo={0};
						//Read out the gauge data
						DWORD sizeofstruct=sizeof(RBRGaugeInfo);
						if(ReadProcessMemory(game_process_handle, (LPCVOID)baseaddress, &gaugeinfo, sizeofstruct, &rw))
						{
							if(rw == sizeofstruct)
							{
								//Read out the car information
								if(ReadProcessMemory(game_process_handle, (LPCVOID)9369184, &baseaddress, sizeof(unsigned long), &rw)) //Base memory location where the next memory address is stored
								{
									if(rw==sizeof(unsigned long) && baseaddress != 0 && baseaddress < 0xFFFFFF00)
									{
										sizeofstruct=sizeof(RBRCarInfo);
										if(ReadProcessMemory(game_process_handle, (LPCVOID)baseaddress, &carinfo, sizeofstruct, &rw))
										{
											if(rw == sizeofstruct)
											{
												//Check if the game is in pause mode or timed out for some reason
												if(oldracetime == gaugeinfo.racetime)
												{
													if(outcount > 4)
													{
														outarray.data[21]=1;
													}
													else
													{
														outcount++;
													}
												}
												else
												{
													outcount=0;
												}
												//calculate time from previous readout to now in s
												double timeleft=0;
												if(oldracetime != 0 && gaugeinfo.racetime != 0)
												{
													timeleft=gaugeinfo.racetime-oldracetime;
													if(timeleft < 0){timeleft=0;}
												}

												//Convert to Euler degrees (wikipedia: http://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles)
												float rad_pitch	= atan2(2*((carinfo.car_look_at[0]*carinfo.car_look_at[3])+(carinfo.car_look_at[1]*carinfo.car_look_at[2])),1-(2*((carinfo.car_look_at[0]*carinfo.car_look_at[0])+(carinfo.car_look_at[1]*carinfo.car_look_at[1]))));
												float rad_roll  =  asin(2*((carinfo.car_look_at[0]*carinfo.car_look_at[2])-(carinfo.car_look_at[3]*carinfo.car_look_at[1])));
												float rad_yaw	= atan2(2*((carinfo.car_look_at[0]*carinfo.car_look_at[1])+(carinfo.car_look_at[2]*carinfo.car_look_at[3])),1-(2*((carinfo.car_look_at[1]*carinfo.car_look_at[1])+(carinfo.car_look_at[2]*carinfo.car_look_at[2]))));

												//2D transformation of speeds (FAQ of www.x-simulator.de)
												float speedz = -sin(rad_yaw) * carinfo.map_speed_x + cos(rad_yaw) * carinfo.map_speed_y;
												float speedy = carinfo.map_speed_z;
												float speedx =  cos(rad_yaw) * carinfo.map_speed_x + sin(rad_yaw) * carinfo.map_speed_y;

												//Calculate accelerations out of change of speed
												float accelx = ((double(speedx)*1000000.000) - oldspeedx)*timeleft;
												float accely = ((double(speedy)*1000000.000) - oldspeedy)*timeleft;
												float accelz = ((double(speedz)*1000000.000) - oldspeedz)*timeleft;
												oldspeedx=double(speedx)*1000000.000;
												oldspeedy=double(speedy)*1000000.000;
												oldspeedz=double(speedz)*1000000.000;

												//Store current time
												oldracetime=gaugeinfo.racetime;
												//collect the outarray package
												if(outarray.data[21]==0) //game is active
												{
													//Transfer the simulator data
													outarray.data[1 ]=int(gaugeinfo.gear_number);
													outarray.data[2 ]=int(gaugeinfo.speed);

													//Car oriented accelerations
													outarray.data[5]=int(accelx*-1000.000);
													outarray.data[6]=int(accely*-1000.000);
													outarray.data[7]=int(accelz*-1000.000);

													//Output the car orientation in degree
													outarray.data[8 ]=int(rad_pitch*-(180.000/M_PI));
													float yawdegree=rad_yaw*(180.000/M_PI);
													if(yawdegree < 0){yawdegree=360.00+yawdegree;}
													outarray.data[9 ]=int(yawdegree);
													outarray.data[10]=int(rad_roll*-(180.000/M_PI));

													//Car oriented speeds
													outarray.data[11]=int(speedx*-1000.000);
													outarray.data[12]=int(speedy*-1000.000);
													outarray.data[13]=int(speedz*-1000.000);

													//Ingame times
													outarray.data[19]=int(gaugeinfo.racetime*1000.000);
													outarray.data[20]=int(gaugeinfo.sessiontime*1000.000);

													//Transfer the gauge data
													outarray.speed=int(gaugeinfo.speed);
													outarray.tempfuel=int(gaugeinfo.temp);
													outarray.currentgear=int(gaugeinfo.gear_number)-1;
													outarray.rpm=int(gaugeinfo.rpm);
													outarray.position=int(gaugeinfo.way_after);
													outarray.attidude=int(accely*1000.000);
													outarray.turncoordinator=int(accelx*1000.000);
													return outarray;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	//Failed
	outarray.data[21]=1;
	return outarray;
}